/* *******************************************************************

   Switch and Read pins on a ESP8266 / ESP32

   Hardware
   This sketch supports several hardware variants:
   - ESP32 NodeMCU D1 R32
   - H801 PWM Dimmer  ( to be re-implemented)    https://werner.rothschopf.net/microcontroller/202106_esp8266_h801_dimmer_pwm_en.htm --> https://amzn.to/3Br4BS5
   - ESP8266 12F - 8 channel relay board         https://werner.rothschopf.net/microcontroller/202108_esp8266_esp12f_relay_x8_en.htm
   - ESP8266 12F - 16 channel shift register board
   - ESP8266 1 channel board (Sonoff like)       https://werner.rothschopf.net/201807_arduino_esp8266_wifimodule.htm                 --> https://amzn.to/3jrMoxG
   - ESP32 Relay X8                              https://werner.rothschopf.net/microcontroller/202208_esp32_relay_x8_en.htm          --> https://s.click.aliexpress.com/e/_Deq5iJV
   - tailor made NodeMCU (or generic ESP8266)
   - tailor made ESP32

   Software
   The Software was last tested with
   - IDE 1.8.19
   - ESP32 Core 2.0.6, 1.0.6    (2.0.8 can have issues)
   - ESP8266 Core 3.0.2, 2.7.4

   Short
   - simple webserver with some pages
   - Stylesheet (css) optimized for mobile devices
   - webseite update with fetch API
   - control pins via webpage
   - ArduinoOTA upload with Arduino IDE
   - see the full tutorial on: https://werner.rothschopf.net/

   Open Tasks
   - add Sonoff S20
   - EEPROM storage of initial values/random values

   by noiasca

   Version
   2023-07-32 Configuration for Arduino Nano ESP32
   2023-07-31 minor fixes for config1.h
   2023-05-07 New board: ESP12F Relay X16 - experimental (config72.h)
   2022-08-01 New board: ESP32 Relay X8 (config75.h)
   2022-02-05 added some comments to config1.h
   2021-09-18 fixes for ESP32
   2021-08-28 New board: D1 R32 AZ Delivery
   2021-08-19 New board: ESP8266 8 channel relay board
   2021-08-12 added several hardware variants based on the esp8266 H801 sketch from 2021-07-30
***************************************************************** */

#ifdef ARDUINO_ARCH_ESP8266            // libraries if you compile for a ESP8266
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>          // for the webserver
#include <ESP8266mDNS.h>               // Bonjour/multicast DNS, finds the device on network by name
#include <ArduinoOTA.h>                // OTA Upload via ArduinoIDE
#endif
#ifdef ARDUINO_ARCH_ESP32              // libraries if you compile for a ESP32
#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>                 // for the webserver
#include <ESPmDNS.h>                   // Bonjour/multicast DNS, finds the device on network by name   
#include <ArduinoOTA.h>                // OTA Upload via ArduinoIDE
#endif

#define VERSION "23.11"                // the version of this sketch
#include "general.h"                   // some general definitions (needed for the following configuration)

/* *******************************************************************
         define which board variant should be compiled
         set the number to one of your defined boards
 ********************************************************************/

#define USE_BOARD 1                    // the actual board to compile:
//                                         1 a generic demo / base configuration for your boards
//                                        26 Sonoff S26 R2 / Sonoff S20
//                                        32 the Arduino Nano ESP32
//                                        65 is a one channel Relay board
//                                        71 the ESP8266 8 channel Relay Board
//                                        72 ESP12F Relay X16 (shift registers)
//                                        75 the ESP32 8 channel Relay Board
//                                        76 D1 R32 Board from AZ Delivery
//                                        79 H801 PWM Dimmer
//                                        99 is an example with a NodeMCU V2
#include "config.h"                    // this file holds several hardware options don't change this line

/* *******************************************************************
         other settings / weitere Einstellungen für den Anwender
 ********************************************************************/

//#include <credentials.h>               // if you have an external file with your credentials you can use it - remove before upload

#ifndef STASSID                        // either use an external .h file containing STASSID and STAPSK - or 
#define STASSID "your-ssid"            // ... modify these line to your SSID
#define STAPSK  "your-password"        // ... and set your WIFI password
#endif

const char* ssid = STASSID;
const char* password = STAPSK;

/* *******************************************************************
         Globals - Variables and constants
 ********************************************************************/
#include "hardware.h"                  // include hardware after config

unsigned long ss = 0;                  // current second since startup
const uint16_t ajaxIntervall = 5;      // intervall for AJAX/FetchAPI call of website in seconds
uint32_t internalVcc = 0;              // only used for ESP8266 - dummy value on ESP32

#ifdef ARDUINO_ARCH_ESP8266
ADC_MODE(ADC_VCC);                     // to use getVcc
ESP8266WebServer server(80);           // an instance for the webserver
#ifndef NO_PWMPIN
const size_t noOfPwmPin = sizeof(pwmPin) / sizeof(pwmPin[0]);
#endif
#else
WebServer server(80);
#ifndef NO_PWMPIN
const size_t noOfPwmPin = (sizeof(pwmPin) / sizeof(pwmPin[0]) > 15) ? 15 : sizeof(pwmPin) / sizeof(pwmPin[0]);
#endif
#endif

// remove spaces from c-string
void removeSpace(char* s) {
  for (char* s2 = s; *s2; ++s2) {
    if (*s2 != ' ')
      *s++ = *s2;
  }
  *s = 0;
}

//IDE OTA
void ArduinoOTASetup(const char * myhostname) {
#ifdef ARDUINO_ARCH_ESP8266
  (void)myhostname;                    // ESP8266 doesn't need a separate hostname for OTA
#endif
#ifdef ARDUINO_ARCH_ESP32
  ArduinoOTA.setHostname(myhostname);  // give a name to your ESP for the Arduino IDE - not needed for ESP8266
#endif
  ArduinoOTA.begin();                  // OTA Upload via ArduinoIDE https://arduino-esp8266.readthedocs.io/en/latest/ota_updates/readme.html
  // OTA Messages
  ArduinoOTA.onStart([]() {
    Terminal.println("OTA Start");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    static byte previous = 0;
    byte result = progress / (total / 100);
    if (previous != result) {
      if ((result % 10) == 0) {
        Terminal.print(result);
        Terminal.println("%");
      }
      else Terminal.print(".");
    }
    previous = result;
  });
  ArduinoOTA.onEnd([]() {
    Terminal.println("\nOTA transmission end");  // afterwards the ESP will perform the update
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Terminal.printf("\nOTA Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Terminal.println("\nAuth Failed");
    else if (error == OTA_BEGIN_ERROR) Terminal.println("\nBegin Failed");
    else if (error == OTA_CONNECT_ERROR) Terminal.println("\nConnect Failed");
    else if (error == OTA_RECEIVE_ERROR) Terminal.println("\nReceive Failed");
    else if (error == OTA_END_ERROR) Terminal.println("\nEnd Failed");
  });
}

/* *******************************************************************
         S E T U P
 ********************************************************************/

void setup(void) {
  Terminal.begin(115200);
  Terminal.print("\n" TXT_BOARDNAME "\nVersion: " VERSION " Board ");
  Terminal.println(USE_BOARD);
  Terminal.print  (__DATE__);
  Terminal.print  (" ");
  Terminal.println(__TIME__);
  //printDefinitions();// debug only

  // initialize the hardware pins
#ifndef NO_OUTPUTPIN
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++) {
    digitalWrite(outputPin[i].pin, !outputPin[i].ON); // switch off output
    pinMode(outputPin[i].pin, OUTPUT);
  }
#endif
#ifndef NO_PWMPIN
#ifdef ARDUINO_ARCH_ESP32              // the ESP32 will need some additional lines of code to simulate the "analogWrite"
  Terminal.print("PWM channels="); Serial.println(noOfPwmPin);
  for (size_t i = 0; i < noOfPwmPin; i++) {
    ledcSetup(i, 5000, 10);            // channel, frequency, resolution in bit (we use 10 to be compatible to ESP8266)
    ledcAttachPin(pwmPin[i].pin, i);   // assign pin to channel
    ledcWrite(i, pwmPin[i].value);
    // to be tested if we can omit pinMode
  }
#else
  for (size_t i = 0; i < noOfPwmPin; i++) {
    pinMode(pwmPin[i].pin, OUTPUT);
    analogWrite(pwmPin[i].pin, pwmPin[i].value);
  }
#endif
#endif

#ifndef NO_INPUTPIN
  for (auto &i : inputPin) pinMode(i.pin, INPUT_PULLUP);
#endif
#ifndef NO_SHIFTPIN
  shiftBegin();
#endif

  if (wifiActivityPin < 255)
    pinMode(wifiActivityPin, OUTPUT);
  char myhostname[32] {"esp"};         // In this sketch we generate a hostname esp+USE_Board+TXT_Boardname
  char val[8];
  itoa(USE_BOARD, val, 10);            // before we can concatenate the precompiler integer, we have to convert it to char array
  strcat(myhostname, val);
  strcat(myhostname, TXT_BOARDNAME);
  removeSpace(myhostname);             // just in case there are any spaces
  WiFi.hostname(myhostname);
#ifdef ARDUINO_ARCH_ESP32
  WiFi.setSleep(false);                // disable powersafe https://github.com/espressif/arduino-esp32/issues/1484
#endif
  Terminal.print("try ");
  Terminal.println(ssid);
  WiFi.begin(ssid, password);
  // Wait for connection
  int8_t i = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Terminal.print(".");
    if (wifiActivityPin < 255) digitalWrite(wifiActivityPin, !digitalRead(wifiActivityPin)); // blink one pin as indicator
    i++;
    if (i >= 80) {
      Terminal.println();
      i = 0;
    }
  }
  Terminal.print("\nConnected to ");
  Terminal.println(ssid);
  Terminal.print("IP address: ");
  Terminal.println(WiFi.localIP());
  if (MDNS.begin(myhostname)) {
    Terminal.println("MDNS responder started");
    MDNS.addService("http", "tcp", 80);          // Add service to MDNS-SD
  }
  if (wifiActivityPin < 255) digitalWrite(wifiActivityPin, wifiActivityPin);
  //define the pages and other content for the webserver
  server.on("/",      handlePage);               // send root page
  server.on("/0.htm", handlePage);               // a request can reuse another handler
  server.on("/1.htm", handlePage1);              // simpler edit of pins
  server.on("/9.htm", handlePage9);              // output of values
  server.on("/f.css", handleCss);                // a stylesheet
  server.on("/j.js",  handleJs);                 // javscript based on fetch API to update the page
  server.on("/json",  handleJson);               // send data in JSON format
  server.on("/cmd",   handleCommand);            // process commands
  server.on("/favicon.ico", handle204);          // not available, send 204
  server.onNotFound(handleNotFound);             // show a typical HTTP Error 404 page
  server.begin();                                // start the webserver
  Terminal.println("HTTP server started\n");

  ArduinoOTASetup(myhostname);
}

/* *******************************************************************
         M A I N L O O P
 ********************************************************************/

void loop(void) {
  ss = millis() / 1000;
  server.handleClient();               // call the webserver
#ifdef ARDUINO_ARCH_ESP8266
  MDNS.update();                       // not on ESP32
#endif
  ArduinoOTA.handle();                 // OTA Upload via ArduinoIDE
  // do your calculation, sensor readings etc.

  // add displays, serial outputs here
}
//
